
/*
*	Error Codes
*		100 - 199: removeFromOverlays
*		200 - 299: removeFromChrome 
*		300 - 399: removeRdfElement 
*		400 - 499: writeCleanupRdf
*		500 - 599: customInfo
*		600 - 699: findRelatedExtensions
*		700	- 720: startBackup
*		721 - 739: addBackupFile
*		740 - 799: revert
*		1000 - 1099: Custom Extension Function (#uninstallFunc)	
*/

var hasCleaned = false;
	
/**
* Extension Uninstaller API is used to completely remove an extension from Mozilla.
* Initialize the uninstall process by passing in the chrome uri
* @constructor
* @param {string} packageUri The chrome uri for the package to uninstall. For example: <pre>urn:mozilla:package:stockticker</pre>
*/
function ExtensionUninstall( sPackageUri ){	
		
	// Load Paths
	var oProps = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);

	/**
	* @private
	*/
	this.appDir = oProps.get("AChrom", Components.interfaces.nsILocalFile);
	/**
	* @private
	*/
	this.profileDir = oProps.get("UChrm", Components.interfaces.nsILocalFile);
	/**
	* @private
	*/
	this.mozDir = oProps.get("CurProcD", Components.interfaces.nsILocalFile);
	
	/**
	* Path seperator
	* @private
	*/
	this.ps = (this.appDir.path.indexOf("/") > -1) ? "/" : "\\";	
	
	/**
	* @final
	* @private
	*/
	this.type = "ExtensionUninstall";
	
	/**
	* The API version
	* @type string
	* @final
	*/
	this.version = "2.0";
	
	/**
	* Chrome uri of the package to be uninstalled
	* @type string
	*/
	this.packageUri	= sPackageUri;
	
	/**
	* Base directory used to bacup all uninstalled extensions
	* @type string
	*/
	this.backup = this.profileDir.path + this.ps +"uninstall"+ this.ps +"backup";
	
	/**
	* Array of files that have been backed up
	* @type array
	*/
	this.backups = new Array();
	
	/**
	* Set to true if you only want to test the uninstall and see the generated logs without actually uninstalling the application
	* @type boolean
	*/
	this.test = false;
	
	/**
	*  Returns the logging object
	* @type ExtUninstallLogger
	*/
	this.logger	= new ExtUninstallLogger();
	
	/**
	* List of files that will be deleted after you restart your browser
	* @type array
	*/
	this.cleanup = new Array();
	
	/**
	* Cleanup the extension files automatically
	* If set to true, the extension files will be automatically added to the cleanup RDF file to be deleted after restart.
	* Otherwise you will have to call writeCleanupRdf() manually to have them added to the RDF file.
	* This is useful if you want to modify the list before they're added.
	* @type boolean
	*/
	this.autoCleanup = true;
	
	/**
	* If set to false, the API will not remove any of the main application elements from the chrome.rdf and overlays.
	* @type boolean
	*/
	this.removeContent = true;
	
	/**
	* If set to false, the API will not remove any locales.
	* @type boolean
	*/
	this.removeLocale = true;
	
	/**
	* If set to false, the API will not remove any skins.
	* @type boolean
	*/
	this.removeSkin = true;
	
	/**
	* If it's set to false, the API will not process #uninstallInfo and #uninstallFunc defined in the application's content.rdf.
	* @see {@link http://mozmonkey.com/extuninstaller_api/optimize.php}
	* @type boolean
	*/
	this.doCustomInfo = true;
	
	/**
	* The RDF file that contains the files to remove apon restart
	* @final
	*/
	this.cleanupRdf	= this.profileDir.path + this.ps + "extuninstallapi.rdf";
	
	/** 
	* Array of extensions to remove.
	* This is populated at time of uninstall and includes related extensions that must also be uninstalled
	* @type array
	*/
	this.extensions	= new Array(); 
		
	/**
	* Did any errors occur durring uninstallation?
	* return boolean
	*/
	this.hasError = function(){
		return this.logger.hasError();
	}
	
	/**
	* Were any warnings generated durring uninstallation?
	* return boolean
	*/
	this.hasWarning	= function(){
		return this.logger.hasWarning();
	}
	
	/**
	* Uninstall the selected extension.
	* <b>NOTE:</b> This cannot be stopped once it starts
	* return boolen True on success and False on failure
	*/
	this.uninstall = function(){
		this.extensions = new Array(); 
		
		// Uninstall
		var hasError = false;		
		try{
			this.findRelatedExtensions();
			
			// Make backup directory
			if(this.extensions.length > 0){
				this.startBackup();
			}
			
			if(this.hasError()){
				return false;
			}
			
			// Uninstall all extensions
			for(var e = 0; e < this.extensions.length; e++){
			
				hasError = false;
				try{
					if(this.doCustomInfo){
						this.customInfo(this.extensions[e]);
					}
					
					// If no errors logged from customInfo
					if(!this.hasError() && !this.extensions[e].logItem.hasError()){
						
						if(this.removeContent){
							this.removeFromChrome(this.extensions[e]);
							this.removeFromOverlays(this.extensions[e]);
						}
						
						this.addToCleanupList(this.extensions[e].logItem, this.getBaseFile(this.extensions[e].packagePath, this.extensions[e].packageName));
						
						this.logger.addLog(this.extensions[e].logItem);
					}
					else{
						this.extensions[e].logItem.status = this.logger.FAIL_STATUS;
						this.logger.addLog(this.extensions[e].logItem);
						break;
					}
					
				}catch(err){
					hasError = true;
					this.extensions[e].logItem.status = this.logger.FAIL_STATUS;
					this.logger.addLog(this.extensions[e].logItem);
					break;
				}
			}
			
			// Add all cleanup files to the RDF
			if(!this.hasError() && this.autoCleanup){
				this.writeCleanupRdf();
			}
			
			hasError = this.hasError();
		}catch(err){ hasError = true; }
		
		return !hasError;
	}
	
	/**
	* Open the built-in log viewer dialog
	* @param {boolean} isModal If set to true the dialog will be modal
	* @param {object} openFrom If not <b>null</b> it will try to open from that object (window, opener, etc.), i.e <code>openFrom.openDialog(...)</code>
	*/
	this.showLogDialog =  function(isModal, openFrom){
		var sModal = ", modal";
		if(isModal == null || isModal == false)
			sModal = "";
		
		if(openFrom != null && openFrom.openDialog){
			openFrom.openDialog('chrome://extuninstallapi/content/log_dialog.xul','ExtUninstallLogDialog','centerscreen, chrome, resizable'+ sModal, this).focus();
		}
		else if(window){
			window.openDialog('chrome://extuninstallapi/content/log_dialog.xul','ExtUninstallLogDialog','centerscreen, chrome, resizable'+ sModal, this).focus();
		}
		else if(opener){
			opener.openDialog('chrome://extuninstallapi/content/log_dialog.xul','ExtUninstallLogDialog','centerscreen, chrome, resizable'+ sModal, this).focus();
		}
	}
	
	/**
	* Get all extensions that use the same applicate file(s).
	* All these related extensions will also be uninstalled
	* @private
	*/
	this.findRelatedExtensions = function(){
		try{	
			// Define RDF Objects
			var	oRdf 		= Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
			var	oDs			= oRdf.GetDataSourceBlocking("rdf:chrome");
			var oRdfC 		= Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
			var	oRdfUtil	= Components.classes["@mozilla.org/rdf/container-utils;1"].getService(Components.interfaces.nsIRDFContainerUtils);
			var oPackage	= oRdf.GetResource("urn:mozilla:package:root");
			var oBaseRes	= oRdf.GetResource("http://www.mozilla.org/rdf/chrome#baseURL");
			var oNameRes	= oRdf.GetResource("http://www.mozilla.org/rdf/chrome#name");
			
			// Get Base Application Path
			try{
				var sAppPath 	= null;
				var oBasePath 	= oDs.GetTarget(oRdf.GetResource(this.packageUri), oBaseRes, true);
				var oBaseName 	= oDs.GetTarget(oRdf.GetResource(this.packageUri), oNameRes, true);
				if(oBasePath instanceof Components.interfaces.nsIRDFLiteral && oBaseName instanceof Components.interfaces.nsIRDFLiteral){
					
					// Add to array
					oExt = new ExtUninstallItem(oBaseName.Value);
					oExt.packageUri 	= this.packageUri;
					oExt.packagePath 	= oBasePath.Value;
					this.extensions.push(oExt);
					
					// Set App Path
					var iFind	 = -1;
					sAppPath = oBasePath.Value;
					if((iFind = sAppPath.lastIndexOf("!")) > -1){
						sAppPath = sAppPath.substring(0, iFind);
					}
				}
				else{
					this.logger.addLog(new ExtuninstallLogItem(this.logger.READ_ACTION, this.logger.FAIL_STATUS, "Getting Extension Information", 602, "Failed to retrieve extension info from Chrome.rdf"));
					throw 602;
				}
			}catch(err){
				if(isNaN(err)){
					this.logger.addLog(new ExtuninstallLogItem(this.logger.READ_ACTION, this.logger.FAIL_STATUS, "Getting Extension Information", 601, err));
					throw 601;
				}
				else{ throw err; }
			}
			
			// Look through Root Package Sequence
			if(oRdfUtil.IsSeq(oDs, oPackage)){
				oRdfC.Init(oDs, oPackage);
				
				var oRes	= null;
				var oBase	= null;
				var oName	= null;
				var oExt	= null;
				var aElems 	= oRdfC.GetElements();
				while(aElems.hasMoreElements()){
					oRes = aElems.getNext();
					
					// Get #baseURL of this package
					if(oRes instanceof Components.interfaces.nsIRDFResource && oRes.Value != this.packageUri){
						oBase	= oDs.GetTarget(oRes, oBaseRes, true);
						oName	= oDs.GetTarget(oRes, oNameRes, true);
						
						// Match
						if(oName instanceof Components.interfaces.nsIRDFLiteral 
							&& oBase instanceof Components.interfaces.nsIRDFLiteral 
							&& oBase.Value.indexOf(sAppPath) == 0 ){
							
							// Creat Extension Item
							oExt = new ExtUninstallItem(oName.Value);
							oExt.packageUri 	= oRes.Value;
							oExt.packagePath 	= oBase.Value;
							this.extensions[this.extensions.length] = oExt;
						}
					}
				}
			}
		}catch(err){ 
			if(!isNaN(err)) { 
				this.logger.addLog(new ExtuninstallLogItem(this.logger.READ_ACTION, this.logger.FAIL_STATUS, "Getting Related Extensions", 603, err));
				throw err; 
			}
		}
	}
	
	
	/**
	* Get Custom uninstall info from contents.rdf 
	* @param {ExtUninstallItem} oExt Extension to be uninstalled
	* @private
	*/
	this.customInfo = function(oExt){
		var sDsUri 	= "chrome://"+ oExt.packageName +"/content/contents.rdf";
		try{
			
			var oParentLog = new ExtuninstallLogItem(this.logger.READ_ACTION, this.logger.SUCCESS_STATUS, "Uninstall custom info", null, null);
			
			// Define RDF Objects
			var	oRdf 		= Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
			var oDs			= oRdf.GetDataSourceBlocking(sDsUri); // nsIRDFDataSource
			var oRdfC 		= Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
			var	oRdfUtil	= Components.classes["@mozilla.org/rdf/container-utils;1"].getService(Components.interfaces.nsIRDFContainerUtils);
					
			// Call Custom Function
			var sFunc	= null;
			var oTrgt 	= oDs.GetTarget(oRdf.GetResource(oExt.packageUri), oRdf.GetResource("http://www.mozilla.org/rdf/chrome#uninstallFunc"), true);
			if(oTrgt instanceof Components.interfaces.nsIRDFLiteral){
				sFunc = oTrgt.Value;
				
				// Remove Parens
				var iParen = -1;
				if((iParen = sFunc.indexOf("(")) > -1)
					sFunc = sFunc.substring(0, iParen);
				
				// Execute Function
				var oLogger = new ExtUninstallLogger();
				var oExecLog = new ExtuninstallLogItem(this.logger.EXECUTE_ACTION, this.logger.SUCCESS_STATUS, (sFunc + "()"), null, null);
				try{
					// Verify Function 
					var v = eval(sFunc);
					
					// Execute
					eval(sFunc + "(oLogger, this.test)");
					
				}catch(err){ 
					if(isNaN(err)){
						oExecLog.status = this.logger.WARN_STATUS;
						oExecLog.errorCode = 501;
						oExecLog.errorMsg = err;
					}
				}
				
				//Add Child Logs
				for(var i = 0; i < oLogger.logs.length; i++){
					oExecLog.addChild(oLogger.logs[i]);
				}
				
				oParentLog.addChild(oExecLog);
			}
			
			
			// Add #InstallInfo Files to Cleanup
			var oSeq 	= oDs.GetTarget(oRdf.GetResource(oExt.packageUri), oRdf.GetResource("http://www.mozilla.org/rdf/chrome#uninstallInfo"), true);
			if(oSeq instanceof Components.interfaces.nsIRDFResource && oRdfUtil.IsSeq(oDs, oSeq)){
				oRdfC.Init(oDs, oSeq);
				
				var oElem		= null;
				var sPath		= null;
				var oFile		= null;
				var sKey		= null;
				var aElems 		= oRdfC.GetElements();
				var	oFile 		= Components.classes["@mozilla.org/file/local;1"].getService(Components.interfaces.nsILocalFile);
				while(aElems.hasMoreElements()){
					oElem = aElems.getNext();
					
					if(oElem instanceof Components.interfaces.nsIRDFLiteral){
						
						// For Directory Service
						if((iKey = oElem.Value.indexOf(":")) > -1 && oProps != null){
							try{
								// Seperate
								sKey 	= oElem.Value.substring(0, iKey);
								sPath	= oElem.Value.substring(iKey + 1);
								
								// Get Key Value & Make Path
								sKey 	= (oProps.get(sKey, Components.interfaces.nsIFile)).path;
								sPath	= sKey + this.ps + sPath;
								
								// Restrict Harmfull Keys
								if(sKey.toLowerCase() == "appdata" 
									|| sKey.toLowerCase() == "appregf"
									|| sKey.toLowerCase() == "wind"){
									return;
								}
								
								// Add to cleanup
								oFile.initWithPath(sPath);
								if( oFile.exists() && sPath.indexOf("..") < 0){
									this.addToCleanupList(oExt, sPath);
								}
							}catch(err){ }
						}
						// For Relative Paths
						else{
							try{
								// Make Path
								sPath = this.mozDir.path + this.ps + oElem.Value;
								
								// Add to cleanup
								oFile.initWithPath(sPath);
								if( oFile.exists() && sPath.indexOf("..") < 0 ){
									this.addToCleanupList(oExt.logItem, sPath);						
								}
							}catch(err){ }
						}
					}
				}
			}
			
		}catch(err){ }
			
		oExt.logItem.addChild(oParentLog);
	}
	
	
	/**
	* Remove elements from Chrome.rdf
	* @param {ExtUninstallItem} oExt Extension to be uninstalled
	* @private
	*/
	this.removeFromChrome = function(oExt){
		
		var oMainLog = new ExtuninstallLogItem(this.logger.READ_ACTION, null, "Chrome Datasources", null, null);
		try{	
			var aRefs	= new Array();
			
			//Define RDF Objects
			var	oDs		= null;
			var	oRdf 	= Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
			var oRdfC 	= Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
			
			//Search Chrome in Application and Profile
			var oParentLog  = null;
			var aChromeDirs = new Array(this.appDir.path, this.profileDir.path);
			for(var d = 0; d < aChromeDirs.length; d++){
				sChromeDir 	= aChromeDirs[d];
				oParentLog 	= new ExtuninstallLogItem(this.logger.READ_ACTION, null, sChromeDir + this.ps + "chrome.rdf", null, null);
				try{
				
					aRefs		= new Array();
					oDs			= oRdf.GetDataSourceBlocking("file:///"+ sChromeDir.replace(new RegExp("\\\\", "g"), "/") + "/chrome.rdf"); // nsIRDFDataSource
					oDs			= oDs.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource); 
					
					// Find References
					var oRes	= null;
					var aSrc	= null;
					var oSrc	= null;
					var aArchs 	= oDs.ArcLabelsIn(oRdf.GetResource(oExt.packageUri));
					while(aArchs.hasMoreElements()){
						oRes 	= aArchs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
						aSrc 	= oDs.GetSources(oRes, oRdf.GetResource(oExt.packageUri), true);
						
						while(aSrc.hasMoreElements()){
							oSrc = aSrc.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
							aRefs[aRefs.length] = oSrc;
						}
					}
					
					//Remove References			
					for(var i = 0; i < aRefs.length; i++){
						// Main
						if(aRefs[i].Value == "urn:mozilla:package:root" && this.removeContent){
							try{
								//Remove
								oRdfC.Init(oDs, oRdf.GetResource("urn:mozilla:package:root"));
								
								this.addBackupFile(sChromeDir + this.ps + "chrome.rdf", oParentLog);
								
								if(!this.test)
									this.removeRdfElement(oExt.packageUri, oDs, oRdf, oRdfC);
								
								oParentLog.addChild(new ExtuninstallLogItem(this.logger.REMOVE_ACTION, this.logger.SUCCESS_STATUS, oExt.packageUri, null, null));
							}catch(err){ 
								oParentLog.addChild(new ExtuninstallLogItem(this.logger.REMOVE_ACTION, this.logger.FAIL_STATUS, oExt.packageUri, 201, err));
								throw 201;
							}
							
						}
						// Locale
						else if(aRefs[i].Value.indexOf("urn:mozilla:locale:") == 0 && aRefs[i].Value.indexOf(oExt.packageName) > -1 && this.removeLocale){
							try{
								
								//Get Locale Root
								var	sLocaleRoot = aRefs[i].Value.substring(19, (Math.abs(aRefs[i].Value.length - oExt.packageName.length))); //extract skin
									sLocaleRoot = "urn:mozilla:locale:"+ sLocaleRoot +"packages";
								
								//Remove
								oRdfC.Init(oDs, oRdf.GetResource(sLocaleRoot));
								
								this.addBackupFile(sChromeDir + this.ps + "chrome.rdf", oParentLog);
								
								if(!this.test)
									this.removeRdfElement(aRefs[i].Value, oDs, oRdf, oRdfC);
								
								oParentLog.addChild(new ExtuninstallLogItem(this.logger.REMOVE_ACTION, this.logger.SUCCESS_STATUS, aRefs[i].Value, null, null));
							}catch(err){ 
								oParentLog.addChild(new ExtuninstallLogItem(this.logger.REMOVE_ACTION, this.logger.FAIL_STATUS, aRefs[i].Value, 202, err));
								throw 202; 
							}
						}
						// Skin
						else if(aRefs[i].Value.indexOf("urn:mozilla:skin:") == 0 && aRefs[i].Value.indexOf(oExt.packageName) > -1 && this.removeSkin){
							try{
								//Get Skin Root
								var sSkinUri = aRefs[i].Value.substring(17, (Math.abs(aRefs[i].Value.length - oExt.packageName.length))); //extract skin
									sSkinUri = "urn:mozilla:skin:"+ sSkinUri +"packages";
								
								//Remove
								oRdfC.Init(oDs, oRdf.GetResource(sSkinUri));
								
								this.addBackupFile(sChromeDir + this.ps + "chrome.rdf", oParentLog);
								
								if(!this.test)
									this.removeRdfElement(aRefs[i].Value, oDs, oRdf, oRdfC);
								
								oParentLog.addChild(new ExtuninstallLogItem(this.logger.REMOVE_ACTION, this.logger.SUCCESS_STATUS, aRefs[i].Value, null, null));
							}catch(err){ 
								oParentLog.addChild(new ExtuninstallLogItem(this.logger.REMOVE_ACTION, this.logger.FAIL_STATUS, aRefs[i].Value, 203, err));
								throw 203; 
							}
						}
					}
				
					//Flush
					oDs.Flush();
					
					oParentLog.status = this.logger.SUCCESS_STATUS;
				}catch(err){
					if(isNaN(err)){
						oParentLog.status 		= this.logger.FAIL_STATUS;
						oParentLog.errorCode	= 204;
						oParentLog.errorMsg		= err;
					}
				}
				oMainLog.addChild(oParentLog);
			}
			
			oMainLog.status = this.logger.SUCCESS_STATUS;
		}catch(err){ 
			
			oMainLog.status 	= this.logger.FAIL_STATUS;
			
			if(isNaN(err)){
				oMainLog.errorCode	= 200;
				oMainLog.errorMsg	= err;
			}
			
			this.logger.addLog(oMainLog);
			throw 200; 
		}
		
		oExt.logItem.addChild(oMainLog);
	}
	
	
	/**
	* Read the contents.rdf for package sPackage
	* @param {ExtUninstallItem} oExt Extension to be uninstalled
	* @private
	*/
	this.removeFromOverlays = function(oExt){
		var sDsUri 		= "chrome://"+ oExt.packageName +"/content/contents.rdf";
		var aOverlays	= new Array();
		var oLogOverlay = new ExtuninstallLogItem(this.logger.READ_ACTION, this.logger.SUCCESS_STATUS, sDsUri, null, null);
		
		try{
		
			// Define RDF Objects
			var	oRdf 		= Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
			var oDs			= oRdf.GetDataSourceBlocking(sDsUri); // nsIRDFDataSource
				oDs			= oDs.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
			var oRdfC 		= Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
			var oRdfC2 		= Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
			var	oRdfUtil 	= Components.classes["@mozilla.org/rdf/container-utils;1"].getService(Components.interfaces.nsIRDFContainerUtils);
			var oSeq		= oRdf.GetResource("urn:mozilla:overlays");
			
			// Collect Overlays
			try{
			
				if(oRdfUtil.IsSeq(oDs, oSeq)){
					oRdfC.Init(oDs, oSeq);
					
					var oRes		= null;
					var sRdf		= null;
					var aSeq		= null;
					var oElem		= null;
					var aElements	= oRdfC.GetElements();
					while(aElements.hasMoreElements()){
						oRes = aElements.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
						sRdf = this.getOverlayPath(oRes.Value);
						
						oRdfC2.Init(oDs, oRes);
						aSeq = oRdfC2.GetElements();
						while(aSeq.hasMoreElements()){
							oElem = aSeq.getNext();
							
							if(oElem instanceof Components.interfaces.nsIRDFLiteral){
								
								if(typeof(aOverlays[sRdf]) == 'undefined'){
									aOverlays[sRdf] = new Array(oElem.Value);
								}
								else{
									aOverlays[sRdf][aOverlays[sRdf].length] = oElem.Value;
								}
							}
						}
					}
				}
				else{
					oLogOverlay.status 		= this.logger.WARNING_STATUS;
					oLogOverlay.errorCode	= 102;
					oLogOverlay.errorMsg	= "No overlay data";
				}
			}catch(err){ 
				oLogOverlay.status 	= this.logger.FAIL_STATUS;
				oLogOverlay.errorCode	= 101;
				oLogOverlay.errorMsg	= err;
				oLogOverlay.addChild(oLogOverlay);
				oExt.logItem.addChild(oLogOverlay);	
				throw 101;
			}
					
			// Remove from Overlays (Application & Profile)
			var sSearch	= "";
			var oDs 	= null;
			var aArchs	= null;
			var aSrc	= null;
			var aChromeDirs = new Array(this.appDir.path, this.profileDir.path);
			for(var p = 0; p < aChromeDirs.length; p++){ // Application & Profile Dirs
					
				// Loop files
				for(sRdfFile in aOverlays){
					var oParentLog = new ExtuninstallLogItem(this.logger.READ_ACTION, null, aChromeDirs[p] + sRdfFile, null, null);
					try{
						oDs	= oRdf.GetDataSourceBlocking("file:///"+ aChromeDirs[p].replace(new RegExp("\\\\", "g"), "/") + sRdfFile);
						oDs	= oDs.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
						
						// Loop overlay elements for this file
						for(var i = 0; i < aOverlays[sRdfFile].length; i++){
							
							// Find Element
							aArchs = oDs.ArcLabelsIn(oRdf.GetLiteral(aOverlays[sRdfFile][i]));
							while(aArchs.hasMoreElements()){
								oRes 	= aArchs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
								aSrc 	= oDs.GetSources(oRes, oRdf.GetLiteral(aOverlays[sRdfFile][i]), true);
								
								// Get parent element for element
								while(aSrc.hasMoreElements()){
									oSrc = aSrc.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
									
									try{
										//Remove
										this.addBackupFile(aChromeDirs[p] + sRdfFile, oParentLog);
										
										if(!this.test)
											oDs.Unassert(oSrc, oRes, oRdf.GetLiteral(aOverlays[sRdfFile][i]));
										
										oParentLog.addChild(new ExtuninstallLogItem(this.logger.REMOVE_ACTION, this.logger.SUCCESS_STATUS, aOverlays[sRdfFile][i], null, null));
									}catch(err){
										oParentLog.addChild(new ExtuninstallLogItem(this.logger.REMOVE_ACTION, this.logger.FAIL_STATUS, aOverlays[sRdfFile][i], 104, err));
									}
								}
							}
						}
						
						oDs.Flush();
						
						oParentLog.status = this.logger.SUCCESS_STATUS;
					}catch(err){ 
						oParentLog.status 		= this.logger.FAIL_STATUS;
						oParentLog.errorCode	= 102;
						oParentLog.errorMsg		= err;
						
						oLogOverlay.addChild(oParentLog);
						oExt.logItem.addChild(oLogOverlay);	
						throw 102;
					}
					oLogOverlay.addChild(oParentLog);
				}
			}
			
			oExt.logItem.addChild(oLogOverlay);
				
		}catch(err){ 
			if(isNaN(err)){
				oLogOverlay.status 		= this.logger.FAIL_STATUS;
				oLogOverlay.errorCode	= 100;
				oLogOverlay.errorMsg	= err;
				oExt.logItem.addChild(oLogOverlay);	
				throw 100; 
			}
			throw err;
		}
	}
	
	
	/**
	* Revert uninstall.
	* Trys to revert all uninstall actions.
	* @return boolean
	*/
	this.revert = function(){
		try{
			this.logger.addLog(new ExtuninstallLogItem(this.logger.WRITE_ACTION, this.logger.EMPTY_STATUS, "Revert uninstall"));
			
			var oRdf = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
						
			for(var i = 0; i < this.backups.length; i++){
				
				// Get target
				var target = this.backups[i].path;
				if(target.indexOf(this.backup + this.ps + "app") == 0){
					target = target.substring((this.backup + this.ps + "app").length);
					target = this.appDir.path + target;
				}
				else if(target.indexOf(this.backup + this.ps + "profile") == 0){
					target = target.substring((this.backup + this.ps + "profile").length);
					target = this.profileDir.path + target;
				}
				else{
					continue;
				}
				
				// File object				
				var trgt = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
				trgt.initWithPath(target);
				
				
				// Copy
				trgt.clone().remove(false);
				this.backups[i].copyTo(trgt.parent, trgt.leafName);
			}
			
		}catch(e){  }
	}
	
	
	/**
	* Create the backup directory
	* @private
	*/
	this.startBackup = function(){
		try{
			// Create base path
			var d = new Date();
			var backup = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
			backup.initWithPath(this.backup);
			backup.appendRelativePath(this.extensions[0].packageName);
			backup.appendRelativePath(d.getTime());
			backup.create(backup.DIRECTORY_TYPE, 0666);
			
			this.backup = backup.path;
			
			// Create sub folders
			var app = backup.clone();
			app.append("app");
			app.create(backup.DIRECTORY_TYPE, 0666);
			
			var profile = backup.clone();
			profile.append("profile");
			profile.create(backup.DIRECTORY_TYPE, 0666);
			
			// Get revert instructions file
			var oRequest	= new XMLHttpRequest();
				oRequest.open("GET", "chrome://extuninstallapi/content/README.txt", false);
	        	oRequest.send("");
			var sContent	= oRequest.responseText;
			
			// Add paths
			sContent = sContent.replace(/<MozAppDir>/g, this.appDir.path);
			sContent = sContent.replace(/<MozProfileDir>/g, this.profileDir.path);
			sContent = sContent.replace(/\n/g, this.linebreak());
			
			// Write
			backup.appendRelativePath("REVERT.txt");
			backup.create(backup.NORMAL_FILE_TYPE, 0666);
			var oFileOut = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
				oFileOut.init(backup, 0x20 | 0x02, 0666, null);
				oFileOut.write(sContent, sContent.length);
				oFileOut.flush();
				oFileOut.close();
			
			this.logger.addLog(new ExtuninstallLogItem(this.logger.WRITE_ACTION, this.logger.SUCCESS_STATUS, "backup directory"));
		}catch(e){
			this.backup = null;
			this.logger.addLog(new ExtuninstallLogItem(this.logger.WRITE_ACTION, this.logger.FAIL_STATUS, "backup directory", 700, "Could not create directory: "+ e));
		}
	}
	
	// Line break is different on various platforms check "http://en.wikipedia.org/wiki/Newline" for a full listing.
	this.linebreak = function(){
		var mac= /mac/i.test(navigator.platform);
		var win= /win/i.test(navigator.platform);
		var unix= /lin|unix|x11/i.test(navigator.platform);
		
		if (win)
			return "\r\n";
		else if (mac)
			return "\r";
		else
			return "\n";
	};
	
	/**
	* Add a file to the backup directory
	* @param {string} file Path to the file to copy to the backup dir
	* @param {ExtUninstallLogger} logger Logger to log to
	* @private
	*/
	this.addBackupFile = function(file, logger){
		var target = file;
		
		try{
			// Set target path
			if(target.indexOf(this.appDir.path) == 0){
				target = target.substring(this.appDir.path.length);
				target = this.backup + this.ps + "app" + target;
			}
			else if(target.indexOf(this.profileDir.path) == 0){
				target = target.substring(this.profileDir.path.length);
				target = this.backup + this.ps + "profile" + target;
			}
			else{
				return;
			}
			
			
			// File object
			var source = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
			source.initWithPath(file);
			
			var trgt = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
			trgt.initWithPath(target);
			
			// Alredy backed up?
			if(trgt.exists()){
				return;
			}
			else if(!trgt.parent.exists()){
				trgt.parent.create(trgt.DIRECTORY_TYPE, 0666);
			}
			
			// Copy
			source.copyTo(trgt.parent, trgt.leafName);
			this.backups.push(trgt);
			
			logger.addChild(new ExtuninstallLogItem(this.logger.WRITE_ACTION, this.logger.SUCCESS_STATUS, "file to backup ("+ target +")"));
		}catch(e){
			logger.addChild(new ExtuninstallLogItem(this.logger.WRITE_ACTION, this.logger.FAIL_STATUS, "file to backup ("+ target +")", 721, "Could not copy file to backup dir ("+ target +"): "+ e));
		}
	}
	
	
	/**
	* Add file path to cleanup rdf file
	* @param {ExtuninstallLogItem} oLogItem Log item to add any log info to.  The log info will be added as log children to this item.
	* @param {string} sFilePath File path to add to the cleanup list
	* @private
	*/
	this.addToCleanupList = function(oLogItem, sFilePath){
		try{
			sFilePath = unescape(sFilePath);
			this.cleanup.push(sFilePath);
			
			this.addBackupFile(sFilePath, oLogItem);
			
			oLogItem.addChild(new ExtuninstallLogItem(this.logger.ADD_ACTION, this.logger.SUCCESS_STATUS, "file to cleanup queue ("+ sFilePath +")"));
		}catch(e){ }
	}
	
	/**
	* Add the file paths from the cleanup array to the cleanup RDF file.
	* These files will be deleted upon restart.
	* <b>NOTE:</b> Only call this if autoCleanup is set to FALSE
	*/
	this.writeCleanupRdf = function(){
		try{
			extuninstall_checkRdf(); // Verify RDF Exists
		
			var sCleanupUri = "http://jgillick.nettripper.com/extuninstallapi/cleanup";
			
			// Define RDF Objects
			var	oRdf 		= Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
			var oDs			= oRdf.GetDataSourceBlocking("file:///"+ this.cleanupRdf);
				oDs			= oDs.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource); 
			var oRdfC 		= Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
			var	oRdfUtil 	= Components.classes["@mozilla.org/rdf/container-utils;1"].getService(Components.interfaces.nsIRDFContainerUtils);
			
			// Get or Create Sequence
			try{
				oRdfC.Init(oDs, oRdf.GetResource(sCleanupUri)); //Get
			}catch(err){
				oRdfC = oRdfUtil.MakeSeq(oDs, oRdf.GetResource(sCleanupUri)); //Create
			}
			
			//Add Files to Cleanup list
			for(var i = 0; i < this.cleanup.length; i++){
				if(!this.test){
					try{
						oRdfC.AppendElement(oRdf.GetLiteral(this.cleanup[i]));
						this.logger.addLog(new ExtuninstallLogItem(this.logger.WRITE_ACTION, this.logger.SUCCESS_STATUS, "To Cleanup ("+ this.cleanup[i] +")"));
					}catch(e){
						this.logger.addLog(new ExtuninstallLogItem(this.logger.WRITE_ACTION, this.logger.WARN_STATUS, "To Cleanup ("+ this.cleanup[i] +")", 400, err));
					}
				}
			}
			
			oDs.Flush();
			
		}catch(err){ 
			this.logger.addLog(new ExtuninstallLogItem(this.logger.WRITE_ACTION, this.logger.WARN_STATUS, "Could not add cleanup files to RDF", 400, err));
		}
	}
	
	
	/**
	* Gets the base file or directory of the application.
	* In most cases this is a JAR file
	* @param {string} sUri  The file URI, i.e. jar:resource:/chrome/app.jar!/content/
	* @param {string} sPackage The name of the pacakge
	* @return string
	* @private
	*/
	this.getBaseFile = function(sUri, sPackage){
		var isJar 		= (sUri.substring(0, 3) == "jar");
			sPackage	= sPackage.toLowerCase();
		
		//Remove JAR text
		if(isJar)
			sUri = sUri.substring(4);
		
		//Translate resource
		if(sUri.substring(0, 8) == "resource"){
			sUri = sUri.substring(9);
			
			//Remove /chrome
			if(sUri.substring(0, 7) == this.ps + "chrome")
				sUri = sUri.substring(7);
			
			sUri = this.appDir.path + sUri;
		}
		
		//Remove '!'
		var iFind = 0;
		if((iFind = sUri.indexOf("!")) > -1){
			sUri = sUri.substring(0, (iFind));
		}
		
		// If File, remove directory after the package name from file path
		//	most likely 'content'
		if(!isJar){
			iFind = sUri.toLowerCase().lastIndexOf(sPackage);
			if(iFind > -1){
				sUri = sUri.substring(0, (iFind + sPackage.length + 1));
			}
		}
		
		//Remove file://
		if(sUri.indexOf("file://") == 0){
			sUri = sUri.substring(7);
			
			// Extra slash for windows
			if(this.ps == "\\"){
				sUri = sUri.substring(1);
			}
		}
		
		//Convert Slashes
		sUri = sUri.replace(/[\\\\]/g, this.ps);
		sUri = sUri.replace(new RegExp("[\\/]", "g"), this.ps);
			
		return sUri;
	}
	
	/**
	* Gets the path to the overlay RDF file.
	* This method converts a path like: <pre>chrome://browser/content/browser.xul</pre>
	* into <pre>/overlayinfo/browser/content/overlays.rdf</pre>
	* @param {string} sXul The path defined in contents.rdf
	* @return string
	* @private
	*/
	this.getOverlayPath = function(sXul){
		
		var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
		var Seamonkey = true;
		try {prefs.getCharPref("general.useragent.extra.seamonkey");} catch (e) {Seamonkey = false;}
		
		if (Seamonkey)
			return this.ps + "overlays.rdf";
		
		//Remove chrome://
		if(sXul.indexOf("chrome://") == 0){
			sXul = sXul.substring(9);
		}
		
		//Add overlayinfo
		sXul = this.ps + "overlayinfo"+this.ps+ sXul;
		sXul = sXul.replace(new RegExp("[\\/]", "g"), this.ps);
		
		//Add overlays.rdf
		var iLast = -1;
		if( (iLast = sXul.lastIndexOf(this.ps)) > -1){
			sXul = sXul.substring(0, (iLast + 1));
			sXul += "overlays.rdf";
		}
		
		return sXul;
	}
	
	/**
	* Removes and RDF element with the specified RDF URI
	* @param {string} sUri The RDF URI of the element to remove
	* @param {nsIRDFDataSource} oDs The RDF datasource
	* @param {nsIRDFService} oRdf The RDF service
	* @param {nsIRDFContainer} oRdfC The RDF Container
	*/
	this.removeRdfElement = function(sUri, oDs, oRdf, oRdfC){
		
		try{
			var oRes 		= oRdf.GetResource(sUri);
			
			//Remove All Archs
			// Loop for duplicates
			var aArchs		= null;
			var hasArchs	= true;
			while(hasArchs){
				aArchs 		= oDs.ArcLabelsOut(oRes);
				hasArchs 	= aArchs.hasMoreElements();
				while(aArchs.hasMoreElements()){
					oArch = aArchs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
					
					//Remove
					oDs.Unassert(oRes, oArch, oDs.GetTarget(oRes, oArch, true));
				}
			}
			
			//Remove Element
			oRdfC.RemoveElement(oRes, true);
			
			//Flush
			if(oDs instanceof Components.interfaces.nsIRDFRemoteDataSource)
				oDs.Flush();
				
		}catch(err){ throw err }
	}
}

/**
* Get the Extension Uninstaller API version
*/
function ExtensionUninstall_version(){
	return (new ExtensionUninstall("")).version;
}

/**
* Uninstall Item Object.
* This stores all the package information for an extension to be uninstalled
* @private
* @constructor
*/
function ExtUninstallItem(packageName){

	this.type			= "ExtUninstallItem";
	this.packageName	= packageName; // Chrome Name of the package
	this.packageUri		= null;	// Chrome URI
	this.packagePath	= null; // Jar or Directory Path
	
	this.logger			= new ExtUninstallLogger();
	this.logItem		= new ExtuninstallLogItem(this.logger.UNINSTALL_ACTION, this.logger.SUCCESS_STATUS, this.packageName, null, null);

}

/**
* Verify that the cleanup RDF (extuninstallapi.rdf) exists
* @private
*/
function extuninstall_checkRdf(){
	try{
		
		var oFileLocal	= Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);	
			oFileLocal.initWithPath((new ExtensionUninstall()).cleanupRdf);	
			
		// Doesn't Exist?
		if(!oFileLocal.exists()){
					
			// Read RDF
			var oRequest	= new XMLHttpRequest();
				oRequest.open("GET", "chrome://extuninstallapi/content/extuninstallapi.rdf", false);
	        	oRequest.send("");
			var sContent	= oRequest.responseText;
			
			// Write
			oFileLocal.create(0x00, 0666);
			var oFileOut = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
				oFileOut.init(oFileLocal, 0x20 | 0x02, 00666, null);
				oFileOut.write(sContent, sContent.length);
				oFileOut.flush();
				oFileOut.close();
		}
	
	}catch(err){ }
}

/**
* Call cleanup 4 seconds after browser starts
* @private
*/
function extuninstall_loadCleanup(){
	if(!hasCleaned){
		hasCleaned = true;
		setTimeout("extuninstall_cleanup()", 4000);
	}
}

/**
* Remove files in cleanup RDF file and old backed up files.
* @private
*/
function extuninstall_cleanup(){
	try{
		// Cleanup RDF
		extuninstall_checkRdf(); // Verify RDF Exists
		
		// Define RDF Objects
		var sRdf		= "file:///"+ (new ExtensionUninstall()).cleanupRdf;
		var	oRdf 		= Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
		var oDs			= oRdf.GetDataSourceBlocking(sRdf);
			oDs			= oDs.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource); 
		var oRdfC 		= Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
		var	oRdfUtil 	= Components.classes["@mozilla.org/rdf/container-utils;1"].getService(Components.interfaces.nsIRDFContainerUtils);
		var oSeq		= oRdf.GetResource("http://jgillick.nettripper.com/extuninstallapi/cleanup");
		
		// Get Cleanup Sequence
		if(oRdfUtil.IsSeq(oDs, oSeq)){
			oRdfC.Init(oDs, oSeq);
		}
		else{
			return;
		}
		
		// Loop through files
		var oElem	= null;
		var aElems 	= oRdfC.GetElements();
		var	oFile 	= Components.classes["@mozilla.org/file/local;1"].getService(Components.interfaces.nsILocalFile);
		while(aElems.hasMoreElements()){
			oElem = aElems.getNext();
			
			if(oElem instanceof Components.interfaces.nsIRDFLiteral){
				
				// Remove file
				try{
					oFile.initWithPath(unescape(oElem.Value));
					
					if(oFile.exists() && oFile.isWritable()){
						oFile.remove(false);
					}
				}catch(err){ }
					
				// Remove from Seq
				oRdfC.RemoveElement(oElem, false);
			}
		}
		
		oDs.Flush();
		
		
		// Old Backup Files
		var old = (new Date()).getTime() - (1000 * 60 * 60 * 24 * 7) // 7 days
		var oProps = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
		var backupDir = oProps.get("UChrm", Components.interfaces.nsILocalFile);
		backupDir.append("uninstall");
		backupDir.append("backup");

		// Loop
		if(backupDir.exists()){
		
			var roots = backupDir.directoryEntries;
			var root = null;	
			while(roots.hasMoreElements()){
				try{
					root = roots.getNext().QueryInterface(Components.interfaces.nsIFile);
					
					// Remove?
					if(root.lastModifiedTime <= old){
						root.remove(true);
					}
					// Loop through sub-direcotries
					else{
						var subs = root.directoryEntries;
						var sub = null;
						while(subs.hasMoreElements()){
							try{
								sub = subs.getNext().QueryInterface(Components.interfaces.nsIFile);
								
								// Remove?
								if(sub.lastModifiedTime <= old){
									sub.remove(true);
								}
							}catch(e){}
						}
					}
				}catch(e){}
			}
		}
		
	}catch(err){ }
}